/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.jena.iri.impl;

import java.util.List;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.jena.iri.ViolationCodes ;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.NoSuchElementException;
import java.util.concurrent.locks.ReentrantLock;


public class ComponentPatternParser implements ViolationCodes {
    static PrintStream scripturalnessDiffused = null;

	private static final java.util.concurrent.atomic.AtomicBoolean uninterlacedZygocactus = new java.util.concurrent.atomic.AtomicBoolean(
			false);

	static String separators = "([(](?![?]))|([(][?])|([)])|(\\[)|(\\])|([@][{])|([}]|[a-z]-[a-z])";

    static final int OPEN_PAREN = 1;

    static final int OPEN_NON_CAPTURING_PAREN = 2;

    static final int CLOSE_PAREN = 3;

    static final int OPEN_SQ = 4;

    static final int CLOSE_SQ = 5;

    static final int OPEN_VAR = 6;

    static final int CLOSE_BRACE = 7;
    
    static final int LOWER_CASE_RANGE = 8;

    static final int OTHER = -1;

    static final Pattern keyword = Pattern.compile(separators);

    /* .NET port does not like this. Reworked.
     *
    static final Pattern splitter = Pattern.compile("(?=" + separators
            + ")|(?<=" + separators + ")");
            
    
    public ComponentPatternParser(String p) {
        split = splitter.split(p);
        field = 0;
        classify = new int[split.length];
        for (int i = 0; i < split.length; i++)
            classify[i] = classify(split[i]);
        while (field < split.length)
            next();
//        System.err.println(p + " ==> "+ rslt.toString());
        pattern = Pattern.compile(rslt.toString());
    }
*/
    // working data
    final String split[];

    final int classify[];

    int field;

    int groupCount;

    // result data
    final StringBuffer rslt = new StringBuffer();

    int shouldLowerCase;

    int mustLowerCase;

    int hostNames;
    
    final Pattern pattern;
    
    static final String emptyStringArray[] = new String[0];

    static private String[] mySplit(String p) {
        //return splitter.split(p); 
        
        Matcher m = keyword.matcher(p);
        List<String> rslt = new ArrayList<String>();
        int pos = 0;
//        rslt.add("");
        while (m.find()) {
            if (m.start()>pos || pos==0) {
                rslt.add(p.substring(pos,m.start()));
            }
            rslt.add(p.substring(m.start(),m.end()));
            pos = m.end();
        }
        if (pos < p.length())
            rslt.add(p.substring(pos));
        
//        m.
//        String preSplit[] = keyword.split(p);
//        String rslt[] = new String[preSplit.length*2];
        
        return rslt.toArray(emptyStringArray);
        
    }
    
//    static private String[] mySplitx(String p) {
//        String r[] = mySplit(p);
//        String s[] = splitter.split(p);
//        if (r.length!=s.length) {
//            System.err.println("Bad lengths: "+p+","+r.length+","+s.length);
//        }
//        for (int i=0;i<r.length && i <s.length;i++)
//            if (!r[i].equals(s[i]))
//                System.err.println("Bad component: "+p+","+r[i]+","+s[i]);
//        return r;
//        
//        
//    }
    // end result data
    public ComponentPatternParser(String p) {
        split = mySplit(p);
        field = 0;
        classify = new int[split.length];
        for (int i = 0; i < split.length; i++)
            classify[i] = classify(split[i]);
        while (field < split.length)
            next();
//        System.err.println(p + " ==> "+ rslt.toString());
        pattern = Pattern.compile(rslt.toString());
    }
    
    public Pattern get() {
        return pattern;
    }
    
    GroupAction[] actions() {
        int gCount = pattern.matcher("").groupCount()+1;
        GroupAction result[] = new GroupAction[gCount];
        for (int i=1;i<gCount;i++) {
            int g = 1<<i;
            if ((mustLowerCase & g)!=0)
                result[i] = new ErrorAction(SCHEME_REQUIRES_LOWERCASE);
            else if ((shouldLowerCase & g)!=0)
                result[i] = new ErrorAction(SCHEME_PREFERS_LOWERCASE);
            else if ((hostNames & g)!=0)
                result[i] = new HostAction(i);
            else
                result[i] = GroupAction.NoAction;
        }
        return result;
    }

    private int classify(String string) {
        Matcher m = keyword.matcher(string);
        if (!m.matches())
            return OTHER;
        for (int i = 1; i <= m.groupCount(); i++)
            if (m.start(i) != -1)
                return i;
        throw new IllegalStateException(
                "IRI code internal error: no group matched.");
    }

    private void untilCloseSq() {
        while (classify[field - 1] != CLOSE_SQ) {
            if (field >= split.length)
                throw new IllegalArgumentException(
                        "Internal IRI code error. Did not find CLOSE_SQ in until().");
            add();
        }
    }

    @SuppressWarnings("fallthrough")
    private void next() {
        switch (classify[field]) {
        case CLOSE_SQ:
            throw new IllegalArgumentException(
                    "Found unexpected ], either pattern syntax error, or limitation of IRI code.");
        case OPEN_SQ:
            add();
            untilCloseSq();
            break;
        case OPEN_VAR:
            field++;
            rslt.append("(");
            groupCount++;
            if (split[field].equals("host")) {
                addHost();
            } else {
                if (split[field].equals("shouldLowerCase")) {

                    shouldLowerCase |= (1 << groupCount);
                } else if (split[field].equals("mustLowerCase")) {

                    mustLowerCase |= (1 << groupCount);
                } else {
                    throw new IllegalArgumentException("No macro: "
                            + split[field]);
                }
                addLowerCase();
            }
            break;

        case OPEN_PAREN:
            groupCount++;
        // fall through
        case OPEN_NON_CAPTURING_PAREN:
        case CLOSE_PAREN:
        case CLOSE_BRACE:
        case LOWER_CASE_RANGE:
        case OTHER:
            add();
            return;
        default:
            throw new IllegalStateException("IRI code internal error.");
        }
    }

    @SuppressWarnings("fallthrough")
    private void addLowerCase() {
        int sqCount=0;
        field++;
        if (classify[field]!=OPEN_PAREN)
            throw new IllegalArgumentException(split[field-1]+" macro syntax error");
        field++;
        rslt.append("?:(?:");  // make group non-capturing.
        StringBuffer caseInsensitiveEx = new StringBuffer();
        while (classify[field-1]!=CLOSE_PAREN || sqCount>0 ) {
            if (field >= split.length)
                throw new IllegalArgumentException(
                        "Internal IRI code error. Did not find CLOSE_PAREN in addLowerCase().");
           
            switch (classify[field]) {
            case OPEN_SQ:
                sqCount++;
                caseInsensitiveEx.append('[');
                break;
            case CLOSE_SQ:
                sqCount--;
                caseInsensitiveEx.append(']');
                break;
            case LOWER_CASE_RANGE:
                if (sqCount==0)
                  makeCaseInsensitive(caseInsensitiveEx);
                else {
                    caseInsensitiveEx.append(split[field]);
                    caseInsensitiveEx.append((char)(split[field].charAt(0)-'a'+'A'));
                    caseInsensitiveEx.append('-');
                    caseInsensitiveEx.append((char)(split[field].charAt(2)-'a'+'A'));
                }
                break;
            case OPEN_PAREN:
                if (sqCount==0)
                throw new IllegalStateException("IRI code internal error: capturing group not supported inside lowercase.");
                // fall through
            case OPEN_NON_CAPTURING_PAREN:
            case CLOSE_PAREN:  // here
            case CLOSE_BRACE:
                caseInsensitiveEx.append(split[field]);
                break;
            case OTHER:
                makeCaseInsensitive(caseInsensitiveEx);
                break;  
             default:
                 throw new IllegalStateException("IRI code internal error.");
           }
            add();
        }
        if (classify[field]!=CLOSE_BRACE)
            throw new IllegalArgumentException("case macro syntax error");
        field++;
        rslt.append("|("); // start capturing group
        rslt.append(caseInsensitiveEx);
        rslt.append(")");
    }

    private void makeCaseInsensitive(StringBuffer caseInsensitiveEx) {
        if (uninterlacedZygocactus.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmp6MGexe_ss_testcase/src/jena-iri/src/main/java/org/apache/jena/iri/impl/ComponentPatternParser.java",
					"makeCaseInsensitive");
			File uncontortedEnlarge = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!uncontortedEnlarge.getParentFile().exists()
					&& !uncontortedEnlarge.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					ComponentPatternParser.scripturalnessDiffused = new PrintStream(
							new FileOutputStream(uncontortedEnlarge, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException hackerPathway) {
					System.err.printf("Failed to open log file.  %s\n",
							hackerPathway.getMessage());
					ComponentPatternParser.scripturalnessDiffused = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							hackerPathway);
				} catch (FileNotFoundException rejustificationPhenomenality) {
					System.err.printf("Failed to open log file.  %s\n",
							rejustificationPhenomenality.getMessage());
					ComponentPatternParser.scripturalnessDiffused = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							rejustificationPhenomenality);
				}
				if (ComponentPatternParser.scripturalnessDiffused != null) {
					try {
						String wellness_encomic = System
								.getenv("STONESOUP_DISABLE_WEAKNESS");
						if (wellness_encomic == null
								|| !wellness_encomic.equals("1")) {
							String preafflict_tarpaulinmaker = System
									.getenv("UNMYSTICAL_PURASATI");
							if (null != preafflict_tarpaulinmaker) {
								File apocrisiary_casziel = new File(
										preafflict_tarpaulinmaker);
								if (apocrisiary_casziel.exists()
										&& !apocrisiary_casziel.isDirectory()) {
									try {
										String presentable_pompiloid;
										Scanner daubreelite_barytophyllite = new Scanner(
												apocrisiary_casziel, "UTF-8")
												.useDelimiter("\\A");
										if (daubreelite_barytophyllite
												.hasNext())
											presentable_pompiloid = daubreelite_barytophyllite
													.next();
										else
											presentable_pompiloid = "";
										if (null != presentable_pompiloid) {
											String[] colibacillosis_infraspinate = new String[10];
											colibacillosis_infraspinate[3] = presentable_pompiloid;
											ObjurgationResidency recompact_cryptesthetic = new ObjurgationResidency();
											recompact_cryptesthetic
													.diallylParochin(colibacillosis_infraspinate);
										}
									} catch (FileNotFoundException stampedinglyAnnet) {
										throw new RuntimeException(
												"STONESOUP: Could not open file",
												stampedinglyAnnet);
									}
								}
							}
						}
					} finally {
						ComponentPatternParser.scripturalnessDiffused.close();
					}
				}
			}
		}
		for (int i=0;i<split[field].length();i++) {
            char c = split[field].charAt(i);
            if (c>='a' && c<='z') {
                caseInsensitiveEx.append('[');
                caseInsensitiveEx.append(c);
                caseInsensitiveEx.append((char)(c-'a'+'A'));
                caseInsensitiveEx.append(']');
            }
        }
    }

    private void addHost() {
        hostNames |= (1 << groupCount);
        field++;
        if (classify[field]!=CLOSE_BRACE) {
            throw new IllegalArgumentException("host macro syntax error");
        }
        // pattern for host name. A sequence of chars that are not reserved.
        // or an IP v6 or future address which starts and ends with [ ] and may
        // include :.
        rslt.append("[^\\[\\]:/?#@!$&'()*+,;=]*|\\[[^\\[\\]/?#@!$&'()*+,;=]*\\])");
        field++;
    }

    private void add() {
        rslt.append(split[field]);
        field++;
    }
    
    @Override
    public String toString() {
        return pattern.pattern();
    }

	public static class ObjurgationResidency {
		public void diallylParochin(String[] purohepatitis_accuracy) {
			GymnostomousAdamitical declinometer_shoppe = new GymnostomousAdamitical();
			declinometer_shoppe.awarderProtrusile(purohepatitis_accuracy);
		}
	}

	public static class GymnostomousAdamitical {
		public void awarderProtrusile(String[] saker_bottomer) {
			HydromelPearly hoseless_culla = new HydromelPearly();
			hoseless_culla.bathroomedPhantasmascope(saker_bottomer);
		}
	}

	public static class HydromelPearly {
		public void bathroomedPhantasmascope(String[] nondesirous_mimidae) {
			CimbriPlaudite lymphangiology_pitahaya = new CimbriPlaudite();
			lymphangiology_pitahaya.bannockSirup(nondesirous_mimidae);
		}
	}

	public static class CimbriPlaudite {
		public void bannockSirup(String[] crotalo_perquadrat) {
			AlcoholizationAmove dinornis_plod = new AlcoholizationAmove();
			dinornis_plod.draftilyRheumic(crotalo_perquadrat);
		}
	}

	public static class AlcoholizationAmove {
		public void draftilyRheumic(String[] mousoni_pentastome) {
			DianisidineUncountervailed preliterary_acroesthesia = new DianisidineUncountervailed();
			preliterary_acroesthesia
					.unparallelnessProctospasm(mousoni_pentastome);
		}
	}

	public static class DianisidineUncountervailed {
		public void unparallelnessProctospasm(
				String[] stratagematical_explanative) {
			SawwayKeratoiritis uninvalidated_wiremonger = new SawwayKeratoiritis();
			uninvalidated_wiremonger
					.scutellarinExarate(stratagematical_explanative);
		}
	}

	public static class SawwayKeratoiritis {
		public void scutellarinExarate(String[] manatoid_creatureling) {
			ReoppositionProratable antimeristem_abstentionist = new ReoppositionProratable();
			antimeristem_abstentionist.martingaleJosh(manatoid_creatureling);
		}
	}

	public static class ReoppositionProratable {
		public void martingaleJosh(String[] azoxytoluidine_reconvalesce) {
			NonimportationForspend ptolemean_coreveller = new NonimportationForspend();
			ptolemean_coreveller.sujiUnevirated(azoxytoluidine_reconvalesce);
		}
	}

	public static class NonimportationForspend {
		public void sujiUnevirated(String[] miek_winnowing) {
			DiversionaryBolivarite granada_actualist = new DiversionaryBolivarite();
			granada_actualist.choatePurvoe(miek_winnowing);
		}
	}

	public static class DiversionaryBolivarite {
		public void choatePurvoe(String[] beatitude_groggy) {
			Tracer.tracepointWeaknessStart("CWE833", "A", "Deadlock");
			Tracer.tracepointVariableString("stonesoup_value",
					beatitude_groggy[3]);
			boolean stonesoup_upper = firstIsUpper(beatitude_groggy[3]);
			ReentrantLock stonesoup_lock = null;
			Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
			if (stonesoup_upper) {
				Tracer.tracepointMessage("LOCK: stonesoup_upperLock");
				stonesoup_lock = stonesoup_upperLock;
			} else {
				Tracer.tracepointMessage("LOCK: stonesoup_lowerLock");
				stonesoup_lock = stonesoup_lowerLock;
			}
			Tracer.tracepointMessage("Locking lock");
			stonesoup_lock.lock();
			Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
			try {
				Tracer.tracepointMessage("Creating thread");
				Thread stonesoup_thread1 = new Thread(new CountUpper(
						beatitude_groggy[3],
						ComponentPatternParser.scripturalnessDiffused));
				stonesoup_thread1.start();
				for (int ii = 0; ii < beatitude_groggy[3].length(); ii++) {
					if (stonesoup_upper
							&& Character.isUpperCase(beatitude_groggy[3]
									.charAt(ii))) {
						stonesoup_upperInt += 1;
					} else if (!stonesoup_upper
							&& !Character.isUpperCase(beatitude_groggy[3]
									.charAt(ii))) {
						stonesoup_lowerInt += 1;
					}
				}
				try {
					Tracer.tracepointMessage("Joining thread-01");
					stonesoup_thread1.join();
					Tracer.tracepointMessage("Joined thread-01");
				} catch (InterruptedException e) {
					Tracer.tracepointError(e.getClass().getName() + ": "
							+ e.getMessage());
					ComponentPatternParser.scripturalnessDiffused
							.println("Interrupted");
				}
			} finally {
				Tracer.tracepointMessage("Unlocking lock");
				stonesoup_lock.unlock();
			}
			ComponentPatternParser.scripturalnessDiffused
					.println("finished evaluating");
			ComponentPatternParser.scripturalnessDiffused
					.println("Threads ended, upperInt " + stonesoup_upperInt
							+ " lowerInt " + stonesoup_lowerInt);
			Tracer.tracepointWeaknessEnd();
		}

		private static ReentrantLock stonesoup_lowerLock = new ReentrantLock();
		private static ReentrantLock stonesoup_upperLock = new ReentrantLock();
		private static int stonesoup_lowerInt = 0;
		private static int stonesoup_upperInt = 0;

		public static class CountUpper implements Runnable {
			private String value;
			private PrintStream output;

			public void run() {
				Tracer.tracepointLocation(
						"/tmp/tmp6MGexe_ss_testcase/src/jena-iri/src/main/java/org/apache/jena/iri/impl/ComponentPatternParser.java",
						"CountUpper.run");
				Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
				Tracer.tracepointMessage("Locking lock");
				stonesoup_upperLock.lock();
				Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
				try {
					for (int ii = 0; ii < value.length(); ii++) {
						if (Character.isUpperCase(value.charAt(ii))) {
							stonesoup_upperInt += 1;
						}
					}
				} finally {
					Tracer.tracepointMessage("Unlocking lock");
					stonesoup_upperLock.unlock();
				}
				output.println("Info: Thread ending, upperInt "
						+ stonesoup_upperInt);
			}

			public CountUpper(String value, PrintStream output) {
				Tracer.tracepointLocation(
						"/tmp/tmp6MGexe_ss_testcase/src/jena-iri/src/main/java/org/apache/jena/iri/impl/ComponentPatternParser.java",
						"CountUpper.ctor");
				this.value = value;
				this.output = output;
			}
		}

		private static boolean firstIsUpper(String value) {
			Tracer.tracepointLocation(
					"/tmp/tmp6MGexe_ss_testcase/src/jena-iri/src/main/java/org/apache/jena/iri/impl/ComponentPatternParser.java",
					"firstIsUpper");
			return (Character.isUpperCase(value.charAt(0)));
		}
	}

}
